@@ -8,6 +8,7 @@ from ._models import (
     BaseEndpoint,
     GracefulRetry,
     GracefulThrottle,
+    GracefulValidator,
     GracyConfig,
     GracyRequestContext,
     LogEvent,
@@ -20,7 +21,7 @@ from ._reports._models import GracyAggregatedRequest, GracyAggregatedTotal, Grac
 
 logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")
 
-__version__ = "1.9.1"
+__version__ = "1.10.0"
 
 __all__ = [
     "exceptions",
@@ -30,6 +31,7 @@ __all__ = [
     # Models
     "BaseEndpoint",
     "GracefulRetry",
+    "GracefulValidator",
     "GracyRequestContext",
     "LogEvent",
     "LogLevel",
