@@ -7,8 +7,6 @@ from typing import Any, Callable, Coroutine, Generic, Iterable, cast
 
 import httpx
 
-from gracy._replay._wrappers import read_replay, record_replay_result
-
 from ._configs import custom_config_context, custom_gracy_config
 from ._loggers import (
     DefaultLogMessage,
@@ -27,6 +25,7 @@ from ._models import (
     GracefulRequest,
     GracefulRetry,
     GracefulRetryState,
+    GracefulValidator,
     GracyConfig,
     GracyRequestContext,
     LogEvent,
@@ -35,9 +34,11 @@ from ._models import (
     Unset,
 )
 from ._replay._storages import GracyReplay
+from ._replay._wrappers import read_replay, record_replay_result
 from ._reports._builders import ReportBuilder
 from ._reports._printers import PRINTERS, print_report
-from .exceptions import GracyParseFailed, NonOkResponse, UnexpectedResponse
+from ._validators import AllowedStatusValidator, DefaultValidator, StrictStatusValidator
+from .exceptions import GracyParseFailed
 
 
 async def _gracefully_throttle(controller: ThrottleController, request_context: GracyRequestContext):
@@ -89,10 +90,11 @@ async def _gracefully_retry(
     throttle_controller: ThrottleController,
     request: GracefulRequest,
     request_context: GracyRequestContext,
-    check_func: Callable[[GracyConfig, httpx.Response], bool],
+    validators: list[GracefulValidator],
 ) -> GracefulRetryState:
     retry = cast(GracefulRetry, request_context.active_config.retry)
     state = retry.create_state()
+
     config = request_context.active_config
     result = None
 
@@ -111,8 +113,21 @@ async def _gracefully_retry(
         result = await request()
         report.track(request_context, result)
 
-        state.success = check_func(config, result)
-        failing = not state.success
+        validation_exc: Exception | None = None
+        for validator in validators:
+            try:
+                validator.check(result)
+            except Exception as ex:
+                validation_exc = ex
+                break
+
+        # Even if all validators are passing, we check whether
+        # it should retry for cases like:
+        # e.g. Allow = 404 (so it's a success),
+        #      but Retry it up to 3 times to see whether it becomes 200
+        if config.should_retry(result.status_code, validation_exc) is False:
+            state.success = True
+            failing = False
 
         if retry.log_after:
             process_log_retry(retry.log_after, DefaultLogMessage.RETRY_AFTER, request_context, state, result)
@@ -123,39 +138,6 @@ async def _gracefully_retry(
     return state
 
 
-def _check_strictness(active_config: GracyConfig, result: httpx.Response) -> bool:
-    if active_config.strict_status_code:
-        if not isinstance(active_config.strict_status_code, Unset):
-            strict_statuses = active_config.strict_status_code
-            if not isinstance(strict_statuses, Iterable):
-                strict_statuses = {strict_statuses}
-
-            if HTTPStatus(result.status_code) not in strict_statuses:
-                return False
-
-    return True
-
-
-def _check_allowed(active_config: GracyConfig, result: httpx.Response, strict_pass: bool | None = None) -> bool:
-    if strict_pass is False:
-        # We do this because we don't have to check allowed statuses if strict failed
-        # Otherwise retry logic might be triggered twice
-        # If this is None we assume strict pass either passed or is unset
-        return True
-
-    successful = result.is_success
-    if not successful and active_config.allowed_status_code:
-        if not isinstance(active_config.allowed_status_code, Unset):
-            allowed = active_config.allowed_status_code
-            if not isinstance(allowed, Iterable):
-                allowed = {allowed}
-
-            if HTTPStatus(result.status_code) not in allowed:
-                return False
-
-    return successful
-
-
 def _maybe_parse_result(active_config: GracyConfig, request_context: GracyRequestContext, result: httpx.Response):
     if active_config.parser and not isinstance(active_config.parser, Unset):
         default_fallback = active_config.parser.get("default", UNSET_VALUE)
@@ -196,54 +178,47 @@ async def _gracify(
     if active_config.log_response and isinstance(active_config.log_response, LogEvent):
         process_log_after_request(active_config.log_response, DefaultLogMessage.AFTER, request_context, result)
 
-    strict_pass = _check_strictness(active_config, result)
-    if strict_pass is False:
-        strict_codes: HTTPStatus | Iterable[HTTPStatus] = active_config.strict_status_code  # type: ignore
-        retry_result = None
-        must_break = True
+    validators: list[GracefulValidator] = []
+    if active_config.strict_status_code and not isinstance(active_config.strict_status_code, Unset):
+        validators.append(StrictStatusValidator(active_config.strict_status_code))
+    elif active_config.allowed_status_code and not isinstance(active_config.allowed_status_code, Unset):
+        validators.append(AllowedStatusValidator(active_config.allowed_status_code))
+    else:
+        validators.append(DefaultValidator())
 
-        if active_config.should_retry(result.status_code):
-            retry_result = await _gracefully_retry(
-                report,
-                throttle_controller,
-                request,
-                request_context,
-                check_func=_check_strictness,
-            )
+    if isinstance(active_config.validators, GracefulValidator):
+        validators.append(active_config.validators)
+    elif isinstance(active_config.validators, Iterable):
+        validators += active_config.validators
 
-        if not retry_result or retry_result.failed:
-            if active_config.log_errors and isinstance(active_config.log_errors, LogEvent):
-                process_log_after_request(active_config.log_errors, DefaultLogMessage.ERRORS, request_context, result)
+    validation_exc: Exception | None = None
+    for validator in validators:
+        try:
+            validator.check(result)
+        except Exception as ex:
+            validation_exc = ex
+            break
 
-            if isinstance(active_config.retry, GracefulRetry) and active_config.retry.behavior == "pass":
-                must_break = False
+    must_break = True
+    retry_result: GracefulRetryState | None = None
+    if active_config.should_retry(result.status_code, validation_exc):
+        retry_result = await _gracefully_retry(
+            report,
+            throttle_controller,
+            request,
+            request_context,
+            validators,
+        )
 
-        if must_break:
-            raise UnexpectedResponse(str(result.url), result, strict_codes)
+    if not retry_result or retry_result.failed:
+        if active_config.log_errors and isinstance(active_config.log_errors, LogEvent):
+            process_log_after_request(active_config.log_errors, DefaultLogMessage.ERRORS, request_context, result)
 
-    allowed_pass = _check_allowed(active_config, result, strict_pass=strict_pass)
-    if allowed_pass is False:
-        retry_result = None
-        must_break = True
+        if isinstance(active_config.retry, GracefulRetry) and active_config.retry.behavior == "pass":
+            must_break = False
 
-        if active_config.should_retry(result.status_code):
-            retry_result = await _gracefully_retry(
-                report,
-                throttle_controller,
-                request,
-                request_context,
-                check_func=_check_allowed,
-            )
-
-        if not retry_result or retry_result.failed:
-            if isinstance(active_config.log_errors, LogEvent):
-                process_log_after_request(active_config.log_errors, DefaultLogMessage.ERRORS, request_context, result)
-
-            if isinstance(active_config.retry, GracefulRetry) and active_config.retry.behavior == "pass":
-                must_break = False
-
-            if must_break:
-                raise NonOkResponse(str(result.url), result)
+        if validation_exc and must_break:
+            raise validation_exc
 
     final_result = _maybe_parse_result(active_config, request_context, result)
 
@@ -401,6 +376,7 @@ ANY_COROUTINE = Coroutine[Any, Any, Any]
 def graceful(
     strict_status_code: Iterable[HTTPStatus] | HTTPStatus | None | Unset = UNSET_VALUE,
     allowed_status_code: Iterable[HTTPStatus] | HTTPStatus | None | Unset = UNSET_VALUE,
+    validators: Iterable[GracefulValidator] | GracefulValidator | None | Unset = UNSET_VALUE,
     retry: GracefulRetry | Unset | None = UNSET_VALUE,
     log_request: LOG_EVENT_TYPE = UNSET_VALUE,
     log_response: LOG_EVENT_TYPE = UNSET_VALUE,
@@ -410,6 +386,7 @@ def graceful(
     config = GracyConfig(
         strict_status_code=strict_status_code,
         allowed_status_code=allowed_status_code,
+        validators=validators,
         retry=retry,
         log_request=log_request,
         log_response=log_response,
